'''
Example from section 6.4 in [1]

This script performs internal continuation of the curve x(lambda) defined by

int_{0}^{1}f(a,x,lambda)c(a,x,lambda) d{a}=0

where f(a,x,lambda)=varphi(a;a,x,lambda) is defined by

{varphi'(alpha;a,x,lambda)=g(varphi(alpha;a,x,lambda),a,x,lambda), alpha in [0,a], \\
varphi(0;a,x,lambda)=varphi_0

with g(varphi(alpha;a,x,lambda),gamma(alpha;a,x,lambda),a,x,lambda)=-(a+1)x(lambda^2 + 1)({varphi(alpha;a,x,lambda) - varphi_0}/{x(lambda^2 + 1)(a + 1)} + 1)({gamma(alpha;a,x,lambda) - gamma_0}/{(lambda^2 + 1)(a+1)} +1).

and c(a,x,lambda)=gamma(a;a,x,lambda) is defined by

{gamma'(alpha;a,x,lambda)=h(varphi(alpha;a,x,lambda),a,x,lambda), alpha in [0,a], \\
gamma(0;a,x,lambda)=gamma_0

with h(varphi(alpha;a,x,lambda),gamma(alpha;a,x,lambda),a,x,lambda)=-(a+1)(lambda^2 + 1)({varphi(alpha;a,x,lambda) - varphi_0}/{x(lambda^2 + 1)(a + 1)} + 1)({gamma(alpha;a,x,lambda) - gamma_0}/{(lambda^2 + 1)(a+1)} +1).

Copyright (C) 2019 Alessia Ando

[1] A. Ando, D. Breda, F. Scarabel, Numerical continuation and delay equations: A novel approach
for complex models of structured populations, Discrete Contin. Dyn. Syst. Ser. S, DOI: 110.3934/dcdss.2020165
'''

import numpy as np
from scipy.fftpack import ifft
import time
import copy

#definition of varphi_0 and gamma_0
phi0 = 1.
gamma0 = 1.

#exact expression of x(lambda) to monitor true error
def exact(l):
    global phi0
    global gamma0
    return ((l**2 + 1)*0.5 - gamma0)*phi0/(l**2 + 1)/((l**2 + 1)/3. - gamma0/2)

#collocation on chebyshev extrema in [0,a]
def chebyshev_extrema_01(n):
    return np.array([0.5*(1 + np.cos(k*np.pi/n)) for k in range(n+1)])

#differentiation matrix for chebyshev nodes in [0,a]
def chebdiff_01(n, a):
    f = 1/a
    D = np.zeros((n+1, n+1))
    x = chebyshev_extrema_01(n)
    D[0,0] = (2*n**2 + 1)*f/3.
    D[n,n] = -(2*n**2 + 1)*f/3.
    D[0,n] = (-1)**n*f
    D[n,0] = -(-1)**n*f
    for j in range(1, n):
        D[0,j] = 2*(-1)**j*f/(1 - x[j])
        D[n,j] = -2*(-1)**(n+j)*f/x[j]
    for i in range(1, n):
        D[i,0] = -0.5*(-1)**i*f/(1 - x[i])
        D[i,n] = 0.5*(-1)**(n+i)*f/x[i]
    for i in range(1, n):
        for j in range(1, i):
            D[i,j] = (-1)**(i+j)*f/(x[i] - x[j])
            D[j,i] = -(-1)**(i+j)*f/(x[i] - x[j])
        D[i, i] = -(2*x[i] - 1)*f/(1 - (2*x[i] - 1)**2)
    return D

#Clenshaw-Curtis weights scaled to interval [0,1]
def clencurt_01(n):
    C = np.zeros((n+1,2))
    k = 2*(1+np.arange(np.floor(n/2)))
    C[::2,0] = 2/np.hstack((1, 1-k*k))
    C[1,1] = -n
    V = np.vstack((C,np.flipud(C[1:n,:])))
    F = np.real(ifft(V, n=None, axis=0))
    w = np.hstack((F[0,0],2*F[1:n,0],F[n,0]))
    return 0.5*w

#auxiliary function for f_coll below
def solve_newton_coll(a, l, x, y, N):
    global phi0
    global gamma0
    M = chebdiff_01(N,a)
    M[N,N] = 1.
    for j in range(N):
        M[N,j] = 0.
    expr1 = ((y[:N+1] - phi0)/(x*(l**2 + 1)*(a + 1)) + 1)
    expr2 = ((y[N+1:] - gamma0)/((l**2 + 1)*(a + 1)) + 1)
    func1 = M.dot(y[:N+1]) + x*(l**2 + 1)*(a + 1)*expr1*expr2
    func1[N] = 0
    func2 = M.dot(y[N+1:]) + (l**2 + 1)*(a + 1)*expr1*expr2
    func2[N] = 0
    func = np.concatenate([func1, func2])
    A = np.zeros((2*(N+1), 2*(N+1)))
    A[N, N] = 1.
    A[2*N+1, 2*N+1] = 1.
    ddg = 1/((a + 1)*(l**2 + 1))
    ddf = ddg/x
    for i in range(N):
        for j in range(N+1):
            A[i, j] = M[i, j]
            A[N+1+i, N+1+j] = M[i, j]
        A[i, i] += (a + 1)*x*(l**2 + 1)*ddf*expr2[i]
        A[i, N+1+i] += (a + 1)*x*(l**2 + 1)*expr1[i]*ddg 
        A[N+1+i, N+1+i] += (a + 1)*(l**2 + 1)*expr1[i]*ddg
        A[N+1+i, i] += (a + 1)*(l**2 + 1)*ddf*expr2[i]
    sol = np.linalg.solve(A, -func)
    return sol

#approximation of f by collocation at the initial value (l,x)
def f_coll(a, l, x, N, tol):
    global phi0
    global gamma0
    y = np.array([phi0]*(N+1)+[gamma0]*(N+1))
    sol = solve_newton_coll(a, l, x, y, N)
    norm = np.linalg.norm(sol)
    for i in range(2*N+2):
        y[i] += sol[i]
    it = 0
    while norm > tol:
        it += 1
        sol = solve_newton_coll(a, l, x, y, N)
        norm = np.linalg.norm(sol)
        for i in range(2*N+2):
            y[i] += sol[i]
    return np.concatenate((y[:N],y[N+1:2*N+1]))

#computation of residual
def G(x, l, Phi):
    global phi0
    global gamma0
    nnodes = len(Phi) + 1
    nodes = chebyshev_extrema_01(nnodes-1)
    N = 2*sum([len(phi[0]) for phi in Phi])
    out = np.zeros(N+1)
    fin = np.array([])
    for phi in range(nnodes - 1):
        a = nodes[phi]
        nvars = len(Phi[phi][0]) + 1
        D = chebdiff_01(nvars-1, a)
        diffs1 = D.dot(np.append(Phi[phi][0], phi0))
        diffs2 = D.dot(np.append(Phi[phi][1], gamma0))
        expr1 = ((Phi[phi][0] - phi0)/(x*(l**2 + 1)*(a + 1)) + 1)
        expr2 = ((Phi[phi][1] - gamma0)/((l**2 + 1)*(a + 1)) + 1)
        for i in range(nvars-1):
            out[2*phi*(nvars-1) + i] = diffs1[i] + x*(l**2 + 1)*(a + 1)*expr1[i]*expr2[i]
            out[2*phi*(nvars-1) + i+nvars-1] = diffs2[i] + (l**2 + 1)*(a + 1)*expr1[i]*expr2[i]
        fin = np.append(fin, Phi[phi][0][0]*Phi[phi][1][0])
    fin = np.append(fin, phi0*gamma0)
    weights = clencurt_01(nnodes-1)
    out[N] = fin.dot(weights)
    return out

#auxiliary function for nullspace of a matrix
def null(A, eps=1e-9):
    u, s, vh = np.linalg.svd(A)
    null_mask = (s <= eps)
    null_space = np.compress(null_mask, vh, axis=0)
    return np.transpose(null_space)

def start_direction(x, l, Phi):
    global phi0
    global gamma0
    nnodes = len(Phi) + 1
    nodes = chebyshev_extrema_01(nnodes-1)
    weights = clencurt_01(nnodes-1)
    N = sum([len(phi) for phi in Phi])
    A = np.zeros((N+2, N+2))
    tot = 0
    for phi in range(nnodes - 1):
        a = nodes[phi]
        nvars = len(Phi[phi])/2 + 1
        expr1 = ((Phi[phi][:nvars-1] - phi0)/(x*(l**2 + 1)*(a + 1)) + 1)
        expr2 = ((Phi[phi][nvars-1:] - gamma0)/((l**2 + 1)*(a + 1)) + 1)
        ddg = 1/((a + 1)*(l**2 + 1))
        ddf = ddg/x
        ddx = -(Phi[phi][:nvars-1] - phi0)/(x**2*(l**2 + 1)*(a + 1))
        d1dl = -2*l*(Phi[phi][:nvars-1] - phi0)/(x*(l**2 + 1)**2*(a + 1))
        d2dl = -2*l*(Phi[phi][nvars-1:] - gamma0)/((l**2 + 1)**2*(a + 1))
        D = chebdiff_01(nvars-1, a)
        for i in range(tot, tot + nvars - 1):
            for j in range(tot, tot + nvars - 1):
                A[i,j] = D[i-tot,j-tot]
                A[i+nvars-1,j+nvars-1] = D[i-tot,j-tot]
            A[i, i] += (a + 1)*x*(l**2 + 1)*ddf*expr2[i-tot]
            A[i, nvars-1+i] += (a + 1)*x*(l**2 + 1)*expr1[i-tot]*ddg 
            A[nvars-1+i, nvars-1+i] += (a + 1)*(l**2 + 1)*expr1[i-tot]*ddg
            A[nvars-1+i, i] += (a + 1)*(l**2 + 1)*ddf*expr2[i-tot]
            A[i,N] += (a + 1)*(l**2 + 1)*expr2[i-tot]*(expr1[i-tot] + x*ddx[i-tot])
            A[i,N+1] += (a + 1)*x*(2*l*expr1[i-tot]*expr2[i-tot] + (l**2 + 1)*d1dl[i-tot]*expr2[i-tot] + (l**2 + 1)*expr1[i-tot]*d2dl[i-tot])
            A[i+nvars-1,N] += (a + 1)*(l**2 + 1)*ddx[i-tot]*expr2[i-tot]
            A[i+nvars-1,N+1] += (a + 1)*(2*l*expr1[i-tot]*expr2[i-tot] + (l**2 + 1)*d1dl[i-tot]*expr2[i-tot] + (l**2 + 1)*expr1[i-tot]*d2dl[i-tot])
        A[N,tot] = weights[phi]*Phi[phi][nvars-1]
        A[N,tot+nvars-1] = weights[phi]*Phi[phi][0]
        tot += 2*(nvars - 1)
    d = null(A)
    return d

#computation of Jacobian and update of the variables
def solve_newton(x, l, Phi, b, d):
    global phi0
    global gamma0
    nnodes = len(Phi) + 1
    nodes = chebyshev_extrema_01(nnodes-1)
    weights = clencurt_01(nnodes-1)
    N = sum([len(phi) for phi in Phi])
    A = np.zeros((N+2, N+2))
    out = np.zeros(N+2)
    fin = np.array([])
    tot = 0
    for phi in range(nnodes - 1):
        nvars = len(Phi[phi])/2 + 1
        a = nodes[phi]
        expr1 = ((Phi[phi][:nvars-1] - phi0)/(x*(l**2 + 1)*(a + 1)) + 1)
        expr2 = ((Phi[phi][nvars-1:] - gamma0)/((l**2 + 1)*(a + 1)) + 1)
        ddg = 1/((a + 1)*(l**2 + 1))
        ddf = ddg/x
        ddx = -(Phi[phi][:nvars-1] - phi0)/(x**2*(l**2 + 1)*(a + 1))
        d1dl = -2*l*(Phi[phi][:nvars-1] - phi0)/(x*(l**2 + 1)**2*(a + 1))
        d2dl = -2*l*(Phi[phi][nvars-1:] - gamma0)/((l**2 + 1)**2*(a + 1))
        D = chebdiff_01(nvars-1, a)
        diffs1 = D.dot(np.append(Phi[phi][:nvars-1], phi0))
        diffs2 = D.dot(np.append(Phi[phi][nvars-1:], gamma0))
        for i in range(nvars-1):
            out[2*phi*(nvars-1) + i] = diffs1[i] + x*(l**2 + 1)*(a + 1)*expr1[i]*expr2[i]
            out[2*phi*(nvars-1) + i+nvars-1] = diffs2[i] + (l**2 + 1)*(a + 1)*expr1[i]*expr2[i]
        fin = np.append(fin, Phi[phi][0]*Phi[phi][nvars-1])
        for i in range(tot, tot + nvars - 1):
            for j in range(tot, tot + nvars - 1):
                A[i,j] = D[i-tot,j-tot]
                A[i+nvars-1,j+nvars-1] = D[i-tot,j-tot]
            A[i, i] += (a + 1)*x*(l**2 + 1)*ddf*expr2[i-tot]
            A[i, nvars-1+i] += (a + 1)*x*(l**2 + 1)*expr1[i-tot]*ddg 
            A[nvars-1+i, nvars-1+i] += (a + 1)*(l**2 + 1)*expr1[i-tot]*ddg
            A[nvars-1+i, i] += (a + 1)*(l**2 + 1)*ddf*expr2[i-tot]
            A[i,N] += (a + 1)*(l**2 + 1)*expr2[i-tot]*(expr1[i-tot] + x*ddx[i-tot])
            A[i,N+1] += (a + 1)*x*(2*l*expr1[i-tot]*expr2[i-tot] + (l**2 + 1)*d1dl[i-tot]*expr2[i-tot] + (l**2 + 1)*expr1[i-tot]*d2dl[i-tot])
            A[i+nvars-1,N] += (a + 1)*(l**2 + 1)*ddx[i-tot]*expr2[i-tot]
            A[i+nvars-1,N+1] += (a + 1)*(2*l*expr1[i-tot]*expr2[i-tot] + (l**2 + 1)*d1dl[i-tot]*expr2[i-tot] + (l**2 + 1)*expr1[i-tot]*d2dl[i-tot])
        A[N,tot] = weights[phi]*Phi[phi][nvars-1]
        A[N,tot+nvars-1] = weights[phi]*Phi[phi][0]
        tot += 2*(nvars - 1)
    fin = np.append(fin, phi0*gamma0)
    weights = clencurt_01(nnodes-1)
    out[N] = fin.dot(weights)
    out[N+1] = b
    if b == 1:
        out[:N+1]=np.zeros(N+1)
    for j in range(N+2):
        A[N+1,j] = d[j]
    sol = np.linalg.solve(A, -out)
    return sol

#Newton's method, single continuation step
def pseudo_alc(u0, lam0, Phi, ds, d, tol):
    N = len(d) - 2
    if d[N+1] < 0:
        ds = -ds
    new_u = u0 + ds*d[N]
    new_lam = lam0 + ds*d[N+1]
    tot = 0
    new_Phi = copy.deepcopy(Phi)
    nnodes = len(Phi)
    for phi in range(nnodes):
        nvars = len(Phi[phi])
        for i in range(tot, tot + nvars):
            new_Phi[phi][i-tot] += ds*d[i]
        tot += nvars
    norm = 1.
    it = 0
    while norm > tol and it < 20:
        it += 1
        diff = np.zeros(N)
        tot = 0
        for phi in range(nnodes):
            for i in range(tot, tot + nvars):
                diff[i] = new_Phi[phi][i-tot] - Phi[phi][i-tot]
            tot += nvars
        b = (new_u - u0)*d[N] + (new_lam - lam0)*d[N+1] + diff.dot(d[:N]) - ds
        sol = solve_newton(new_u, new_lam, new_Phi, b, d)
        norm = np.linalg.norm(sol)
        new_u += sol[N]
        new_lam += sol[N+1]
        tot = 0
        for phi in range(nnodes):
            for i in range(tot, tot + nvars):
                new_Phi[phi][i-tot] += sol[i]
            tot += nvars
    new_d = solve_newton(new_u, new_lam, new_Phi, 1., d)
    n = np.linalg.norm(new_d)
    if n != 0:
        new_d = new_d/n
    if it == 20:
        return lam0, u0, Phi, d, it
    return new_lam, new_u, new_Phi, new_d, it

#varying number of collocation points, fixed number of quadrature points
for ncoll in range(7, 8):
    nquad = 10
    l = 2.
    u = exact(l)
    Phi = [f_coll(a, l, u, ncoll-1, 1e-12) for a in chebyshev_extrema_01(nquad)[:nquad]]
    d = start_direction(u, l, Phi)
    ds = 0.1*ncoll
    start = time.time()
    #10 continuation steps
    l, u, Phi, d, it = pseudo_alc(u, l, Phi, ds, d, 1e-13)
    while it == 20:
        ds = ds*0.5
        l, u, Phi, d, it = pseudo_alc(u, l, Phi, ds, d, 1e-13)
    lambdas = np.array([l])
    us = np.array([u])
    its = np.array([it])
    if it < 3:
        ds = ds*1.3
    
    for i in range(9):
        l, u, Phi, d, it = pseudo_alc(u, l, Phi, ds, d, 1e-13)
        while it == 20:
            ds *= 0.5
            l, u, Phi, d, it = pseudo_alc(u, l, Phi, ds, d, 1e-13)
        lambdas = np.append(lambdas, l)
        us = np.append(us, u)
        its = np.append(its, it)
        if it < 3:
            ds = ds*1.3

    elapsed = time.time() - start
    print 'Elapsed time =', elapsed

    errs = []
    for l1, u1, it in zip(lambdas, us, its):
        print l1, abs(u1-exact(l1)), it
        errs.append(abs(u1-exact(l1)))
