'''
Example from section 6.5 in [1]

This script performs internal continuation of the curve S(mu) defined by

int_{a_mat}^{a_max}beta(xi(a,S),S)F(a,S) d{a}=1

where xi(a,S)=bar{xi}(a;a,S) is defined by

{bar{xi}'(alpha;a,S)=g(bar{xi}(alpha;a,S),a,S), alpha in [0,a], \\
bar{xi}(0;a,S)=xi_b,

F(a,mu)=bar{F}(a;a,mu) is defined by

{bar{F}'(alpha;a,mu)=-mu bar{F}(alpha;a,mu), alpha in [0,a], \\
bar{F}(0;a,mu)=1,

a_mat is defined by bar{xi}(a_mat,S)=xi_A,

and a_mat, beta, g, mu, xi_b, xi_A are defined as in [1, Table 1].

Copyright (C) 2019 Alessia Ando

[1] A. Ando, D. Breda, F. Scarabel, Numerical continuation and delay equations: A novel approach
for complex models of structured populations, Discrete Contin. Dyn. Syst. Ser. S, DOI: 110.3934/dcdss.2020165
'''

import numpy as np
from scipy.optimize import fsolve
from scipy.fftpack import ifft
import scipy.integrate as integrate
import time
import copy

#definition of RHS for xi NOT QUITE
def g(xi, alpha, x):
    return 0.006*(300*x/(x+1.5e-5) - xi)

#auxiliary function to detect a_mat at the initial value S=x
def f(a, x):
    Phi, info = integrate.odeint(g, 0.8, np.linspace(0, a[0], 2), args = (x,), full_output = 1, atol = 1e-13, rtol = 1e-13)
    return Phi[-1][0] - 2.5

#exact expression of int_{a_mat}^{a_max}beta(xi(a,S),S)F(a,S) d{a}-1, to monitor error
def error(l, x):
    f = 7*x/(1 + 7*x)
    lg = (0.8 - 6*f)/(2.5 - 6*f)
    e1 = lg**(-(0.3 + l)/0.15) - np.exp(-(0.3 + l)*70)
    e2 = lg**(-(0.15 + l)/0.15) - np.exp(-(0.15 + l)*70)
    e3 = lg**(-l/0.15) - np.exp(-l*70)
    val = 0.1*f*((0.8 - 6*f)**2/(0.3 + l)*e1 + 12*f*(0.8 - 6*f)/(0.15 + l)*e2 + (6*f)**2/l*e3)
    return abs(val - 1)

#collocation on chebyshev extrema in [0,a]
def chebyshev_extrema_01(n):
    return np.array([0.5*(1 + np.cos(k*np.pi/n)) for k in range(n+1)])

#differentiation matrix for chebyshev nodes in [0,a]
def chebdiff_01(n, a):
    f = 1/a
    D = np.zeros((n+1, n+1))
    x = chebyshev_extrema_01(n)
    D[0,0] = (2*n**2 + 1)*f/3.
    D[n,n] = -(2*n**2 + 1)*f/3.
    D[0,n] = (-1)**n*f
    D[n,0] = -(-1)**n*f
    for j in range(1, n):
        D[0,j] = 2*(-1)**j*f/(1 - x[j])
        D[n,j] = -2*(-1)**(n+j)*f/x[j]
    for i in range(1, n):
        D[i,0] = -0.5*(-1)**i*f/(1 - x[i])
        D[i,n] = 0.5*(-1)**(n+i)*f/x[i]
    for i in range(1, n):
        for j in range(1, i):
            D[i,j] = (-1)**(i+j)*f/(x[i] - x[j])
            D[j,i] = -(-1)**(i+j)*f/(x[i] - x[j])
        D[i, i] = -(2*x[i] - 1)*f/(1 - (2*x[i] - 1)**2)
    return D

#Clenshaw-Curtis weights scaled to interval [0,1]
def clencurt_01(n):
    C = np.zeros((n+1,2))
    k = 2*(1+np.arange(np.floor(n/2)))
    C[::2,0] = 2/np.hstack((1, 1-k*k))
    C[1,1] = -n
    V = np.vstack((C,np.flipud(C[1:n,:])))
    F = np.real(ifft(V, n=None, axis=0))
    w = np.hstack((F[0,0],2*F[1:n,0],F[n,0]))
    return 0.5*w

#auxiliary function for f_coll below
def solve_newton_coll(a, l, x, y, N):
    M = chebdiff_01(N,a)
    M[N,N] = 1.
    for j in range(N):
        M[N,j] = 0.
    expr1 = 0.15*(42*x/(1 + 7*x) - y[0])
    expr2 = -l*y[1]
    func1 = M.dot(y[0]) - expr1
    func1[N] = 0
    func2 = M.dot(y[1]) - expr2
    func2[N] = 0
    func = np.concatenate([func1, func2])
    A = np.zeros((2*(N+1), 2*(N+1)))
    A[N, N] = 1.
    A[2*N+1, 2*N+1] = 1.
    for i in range(N):
        for j in range(N+1):
            A[i, j] = M[i, j]
            A[N+1+i, N+1+j] = M[i, j]
        A[i, i] += 0.15
        A[N+1+i, N+1+i] += l
    sol = np.linalg.solve(A, -func)
    return sol

#approximation of f by collocation at the initial value (l,x)
def f_coll(a, l, x, N, tol):
    y = [np.array([0.8]*(N+1)), np.array([1.]*(N+1))]
    sol = solve_newton_coll(a, l, x, y, N)
    norm = np.linalg.norm(sol)
    for i in range(N+1):
        y[0][i] += sol[i]
        y[1][i] += sol[N+1+i]
    it = 0
    while norm > tol:
        it += 1
        sol = solve_newton_coll(a, l, x, y, N)
        norm = np.linalg.norm(sol)
        for i in range(N+1):
            y[0][i] += sol[i]
            y[1][i] += sol[N+1+i]
    return [y[0][:N], y[1][:N]]

#computation of residual
def G(x, l, Phi, age):
    nnodes = len(Phi)
    nodes = (70 - age)*chebyshev_extrema_01(nnodes-1) + age
    N = 2*sum([len(phi[0]) for phi in Phi])
    out = np.zeros(N+2)
    fin = np.array([])
    for phi in range(nnodes):
        a = nodes[phi]
        nvars = len(Phi[phi][0]) + 1
        D = chebdiff_01(nvars-1, a)
        diffs1 = D.dot(np.append(Phi[phi][0], 0.8))
        diffs2 = D.dot(np.append(Phi[phi][1], 1.))
        expr1 = 0.15*(42*x/(1 + 7*x) - Phi[phi][0])
        expr2 = -l*Phi[phi][1]
        for i in range(nvars-1):
            out[2*phi*(nvars-1) + i] = diffs1[i] - expr1[i]
            out[2*phi*(nvars-1) + i+nvars-1] = diffs2[i] - expr2[i]
        fin = np.append(fin, 0.1*(7*x/(1+7*x))*Phi[phi][0][0]**2*Phi[phi][1][0])
    weights = clencurt_01(nnodes-1)*(70 - age)
    out[N] = fin.dot(weights) - 1
    out[N+1] = Phi[nnodes-1][0][0] - 2.5
    return out

#auxiliary function for nullspace of a matrix
def null(A, eps=1e-9):
    u, s, vh = np.linalg.svd(A)
    null_mask = (s <= eps)
    null_space = np.compress(null_mask, vh, axis=0)
    return np.transpose(null_space)

def start_direction(x, l, Phi, age):
    nnodes = len(Phi)
    nodes = chebyshev_extrema_01(nnodes-1)
    weights = clencurt_01(nnodes-1)
    N = 2*sum([len(phi[0]) for phi in Phi])
    A = np.zeros((N+3, N+3))
    tot = 0
    for phi in range(nnodes):
        a = (70 - age)*nodes[phi] + age
        nvars = len(Phi[phi][0]) + 1
        expr1 = 0.15*(42*x/(1 + 7*x) - Phi[phi][0])
        expr2 = -l*Phi[phi][1]
        D = chebdiff_01(nvars-1, a)
        diffs1 = D.dot(np.append(Phi[phi][0], 0.8))
        diffs2 = D.dot(np.append(Phi[phi][1], 1.))
        for i in range(tot, tot + nvars - 1):
            for j in range(tot, tot + nvars - 1):
                A[i,j] = D[i-tot,j-tot]
                A[i+nvars-1,j+nvars-1] = D[i-tot,j-tot]
            A[i, i] += 0.15
            A[nvars-1+i, nvars-1+i] += l
            A[i,N+1] += -6.3/(1+7*x)**2
            A[i+nvars-1,N+2] += Phi[phi][1][i-tot]
            A[i,N] += (nodes[phi] - 1)*diffs1[i-tot]/a
            A[i+nvars-1,N] += (nodes[phi] - 1)*diffs2[i-tot]/a
        A[N,tot] += (70 - age)*weights[phi]*0.2*(7*x/(1+7*x))*Phi[phi][0][0]*Phi[phi][1][0]
        A[N,tot+nvars-1] += (70 - age)*weights[phi]*0.1*(7*x/(1+7*x))*Phi[phi][0][0]**2
        A[N,N+1] += (70 - age)*weights[phi]*0.1*(7/(1+7*x)**2)*Phi[phi][0][0]**2*Phi[phi][1][0]
        tot += 2*(nvars - 1)
    A[N,N] = -sum([weights[phi]*0.1*(7*x/(1+7*x))*Phi[phi][0][0]**2*Phi[phi][1][0] for phi in range(nnodes)])
    A[N+1, 2*(nvars-1)*(nnodes-1)] = 1.
    d = null(A)
    return d

#computation of Jacobian and update of the variables
def solve_newton(x, l, Phi, age, arr, d):
    nnodes = len(Phi)
    nodes = chebyshev_extrema_01(nnodes-1)
    weights = clencurt_01(nnodes-1)
    N = 2*sum([len(phi[0]) for phi in Phi])
    A = np.zeros((N+3, N+3))
    tot = 0
    for phi in range(nnodes):
        a = (70 - age)*nodes[phi] + age
        nvars = len(Phi[phi][0]) + 1
        expr1 = 0.15*(42*x/(1 + 7*x) - Phi[phi][0])
        expr2 = -l*Phi[phi][1]
        D = chebdiff_01(nvars-1, a)
        diffs1 = D.dot(np.append(Phi[phi][0], 0.8))
        diffs2 = D.dot(np.append(Phi[phi][1], 1.))
        for i in range(tot, tot + nvars - 1):
            for j in range(tot, tot + nvars - 1):
                A[i,j] = D[i-tot,j-tot]
                A[i+nvars-1,j+nvars-1] = D[i-tot,j-tot]
            A[i, i] += 0.15
            A[nvars-1+i, nvars-1+i] += l
            A[i,N+1] += -6.3/(1+7*x)**2
            A[i+nvars-1,N+2] += Phi[phi][1][i-tot]
            A[i,N] += (nodes[phi] - 1)*diffs1[i-tot]/a
            A[i+nvars-1,N] += (nodes[phi] - 1)*diffs2[i-tot]/a
        A[N,tot] += (70 - age)*weights[phi]*0.2*(7*x/(1+7*x))*Phi[phi][0][0]*Phi[phi][1][0]
        A[N,tot+nvars-1] += (70 - age)*weights[phi]*0.1*(7*x/(1+7*x))*Phi[phi][0][0]**2
        A[N,N+1] += (70 - age)*weights[phi]*0.1*(7/(1+7*x)**2)*Phi[phi][0][0]**2*Phi[phi][1][0]
        tot += 2*(nvars - 1)
    A[N,N] = -sum([weights[phi]*0.1*(7*x/(1+7*x))*Phi[phi][0][0]**2*Phi[phi][1][0] for phi in range(nnodes)])
    A[N+1, 2*(nvars-1)*(nnodes-1)] = 1.
    for j in range(N+3):
        A[N+2,j] = d[j]
    sol = np.linalg.solve(A, arr)
    return sol

#Newton's method, single continuation step
def pseudo_alc(u0, lam0, Phi, a0, ds, d, tol):
    N = len(d) - 3
    if d[N+2] < 0:
        ds = -ds
    new_u = u0 + ds*d[N+1]
    new_lam = lam0 + ds*d[N+2]
    new_a = a0 + ds*d[N]
    tot = 0
    new_Phi = copy.deepcopy(Phi)
    nnodes = len(Phi)
    for phi in range(nnodes):
        nvars = len(Phi[phi][0])
        for i in range(tot, tot + nvars):
            new_Phi[phi][0][i-tot] += ds*d[i]
            new_Phi[phi][1][i-tot] += ds*d[i+nvars]
        tot += 2*nvars
    norm = 1.
    it = 0
    while norm > tol and it < 20:
        it += 1
        diff = np.zeros(N)
        tot = 0
        for phi in range(nnodes):
            for i in range(tot, tot + nvars):
                diff[i] = new_Phi[phi][0][i-tot] - Phi[phi][0][i-tot]
                diff[i+nvars] = new_Phi[phi][1][i-tot] - Phi[phi][1][i-tot]
            tot += 2*nvars
        b = -np.append(G(new_u, new_lam, new_Phi, new_a), (new_u - u0)*d[N+1] + (new_lam - lam0)*d[N+2] + (new_a - a0)*d[N] + diff.dot(d[:N]) - ds)
        sol = solve_newton(new_u, new_lam, new_Phi, new_a, b, d)
        norm = np.linalg.norm(sol)
        new_u += sol[N+1]
        new_lam += sol[N+2]
        new_a += sol[N]
        tot = 0
        for phi in range(nnodes):
            for i in range(tot, tot + nvars):
                new_Phi[phi][0][i-tot] += sol[i]
                new_Phi[phi][1][i-tot] += sol[i+nvars]
            tot += 2*nvars
    new_d = solve_newton(new_u, new_lam, new_Phi, new_a, np.append(np.zeros(N+2), 1.), d)
    n = np.linalg.norm(new_d)
    if n != 0:
        new_d = new_d/n
    if it == 20:
        return lam0, u0, Phi, a0, d, it
    print new_lam, new_u, new_a, it
    return new_lam, new_u, new_Phi, new_a, new_d, it

#varying number of collocation points, fixed number of quadrature points
for ncoll in range(10,11):
    nquad = 20
    #l = 0.243845788916114
    #x = 0.351318582230538
    #l = 0.0386123306848
    #x = 0.102343563326
    l = 0.001688868564312
    x = 0.126848862788998
    abar = fsolve(f, 4., args = (x,))
    Phi = [f_coll(a, l, x, ncoll-1, 1e-13) for a in (70 - abar)*chebyshev_extrema_01(nquad) + abar]
    d = start_direction(x, l, Phi, abar)
    ds = 0.15*ncoll
    #10 continuation steps
    start = time.time()
    l, u, Phi, a, d, it, = pseudo_alc(x, l, Phi, abar, ds, d, 1e-13)
    while it == 20:
        ds = ds*0.5
        l, u, Phi, a, d, it = pseudo_alc(x, l, Phi, abar, ds, d, 1e-13)
    lambdas = np.array([l])
    us = np.array([u])
    its = np.array([it])
    if it < 3:
        ds = ds*1.3
    
    for i in range(9):
        l, u, Phi, a, d, it = pseudo_alc(u, l, Phi, a, ds, d, 1e-13)
        while it == 20:
            ds *= 0.5
            l, u, Phi, a, d, it = pseudo_alc(u, l, Phi, a, ds, d, 1e-13)
        lambdas = np.append(lambdas, l)
        us = np.append(us, u)
        its = np.append(its, it)
        if it < 3:
            ds = ds*1.3

    elapsed = time.time() - start
    print 'Elapsed time =', elapsed

    errs = []
    for l1, u1, it in zip(lambdas, us, its):
        errs.append(error(l1, u1))
        print l1, u1, error(l1, u1), it
