'''
Example from section 6.3 in [1]

This script performs internal continuation of the curve x(lambda) defined by

int_{a_mat}^{1}f(a,x,lambda) d{a}=0

where f(a,x,lambda)=varphi(a;a,x,lambda) is defined by

{varphi'(alpha;a,x,lambda) = g_{1}(varphi(alpha;a,x,\lambda),a,x,\lambda), alpha in[0,a],\\
varphi(0;a,x,lambda) = varphi_0

if a <= a_mat, while if a > a_mat,

{varphi'(alpha;a,x,lambda) = g_{2}(varphi(alpha;a,x,\lambda),a,x,\lambda), alpha in[a_mat,a],\\
varphi(a_mat;a,x,lambda) = varphi_0/2,\\
varphi'(alpha;a,x,lambda) = g_{1}(varphi(alpha;a,x,\lambda),a,x,\lambda), alpha in[0,a_mat],\\
varphi(0;a,x,lambda) = varphi_0.

with g_{1} = (varphi(alpha;a,x,lambda),a,x,lambda)=-x(lambda^2 + 1)({varphi(alpha;a,x,lambda) - varphi_0}/{x(lambda^2 + 1)} - 1/2)^2, and
g_{2} = (varphi(alpha;a,x,lambda),a,x,lambda)=-(a + 1)x(lambda^2 + 1)({varphi(alpha;a,x,lambda) - varphi_0/2}/{x(lambda^2 + 1)(a + 1)} + 1/{a_mat+1})^2

and a_mat is defined by f(a_mat,x,lambda)=varphi_0/2.

Copyright (C) 2019 Alessia Ando

[1] A. Ando, D. Breda, F. Scarabel, Numerical continuation and delay equations: A novel approach
for complex models of structured populations, Discrete Contin. Dyn. Syst. Ser. S, DOI: 110.3934/dcdss.2020165
'''

import numpy as np
from scipy.fftpack import ifft
import time
import copy

#definition of varphi_0
phi0 = 1.

#exact expression of x(lambda) to monitor true error
def exact(l):
    global phi0
    return 3*phi0/(l**2 + 1)

#collocation on chebyshev extrema in [0,a]
def chebyshev_extrema_01(n):
    return np.array([0.5*(1 + np.cos(k*np.pi/n)) for k in range(n+1)])

#differentiation matrix for chebyshev nodes in [0,a]
def chebdiff_01(n, a):
    f = 1/a
    D = np.zeros((n+1, n+1))
    x = chebyshev_extrema_01(n)
    D[0,0] = (2*n**2 + 1)*f/3.
    D[n,n] = -(2*n**2 + 1)*f/3.
    D[0,n] = (-1)**n*f
    D[n,0] = -(-1)**n*f
    for j in range(1, n):
        D[0,j] = 2*(-1)**j*f/(1 - x[j])
        D[n,j] = -2*(-1)**(n+j)*f/x[j]
    for i in range(1, n):
        D[i,0] = -0.5*(-1)**i*f/(1 - x[i])
        D[i,n] = 0.5*(-1)**(n+i)*f/x[i]
    for i in range(1, n):
        for j in range(1, i):
            D[i,j] = (-1)**(i+j)*f/(x[i] - x[j])
            D[j,i] = -(-1)**(i+j)*f/(x[i] - x[j])
        D[i, i] = -(2*x[i] - 1)*f/(1 - (2*x[i] - 1)**2)
    return D

#Clenshaw-Curtis weights scaled to interval [0,1]
def clencurt_01(n):
    C = np.zeros((n+1,2))
    k = 2*(1+np.arange(np.floor(n/2)))
    C[::2,0] = 2/np.hstack((1, 1-k*k))
    C[1,1] = -n
    V = np.vstack((C,np.flipud(C[1:n,:])))
    F = np.real(ifft(V, n=None, axis=0))
    w = np.hstack((F[0,0],2*F[1:n,0],F[n,0]))
    return 0.5*w

#auxiliary function for f_coll1 below 
def solve_newton_coll1(a, abar, l, x, y, N):
    global phi0
    if abar == 0:
        return np.zeros(N+1)
    M = chebdiff_01(N,abar)
    M[N,N] = 1.
    for j in range(N):
        M[N,j] = 0.
    expr = (y - phi0)/(x*(l**2 + 1)) - 0.5
    func = M.dot(y) + x*(l**2 + 1)*expr**2
    func[N] = 0
    A = np.copy(M)
    ddg = 1/(l**2 + 1)
    ddf = ddg/x
    for i in range(N):
        A[i, i] += 2*x*(l**2 + 1)*ddf*expr[i]
    sol = np.linalg.solve(A, -func)
    return sol

#auxiliary function for f_coll2 below
def solve_newton_coll2(a, abar, l, x, y, N):
    global phi0
    M = chebdiff_01(N,a-abar)
    M[N,N] = 1.
    for j in range(N):
        M[N,j] = 0.
    expr = ((y - phi0*0.5)/(x*(l**2 + 1)*(a + 1)) + 1/(abar+1))
    func = M.dot(y) + x*(l**2 + 1)*(a + 1)*expr**2
    func[N] = 0
    A = np.copy(M)
    ddg = 1/((a + 1)*(l**2 + 1))
    ddf = ddg/x
    for i in range(N):
        A[i, i] += 2*(a + 1)*x*(l**2 + 1)*ddf*expr[i]
    sol = np.linalg.solve(A, -func)
    return sol

#approximation of f by collocation at the initial value (l,x), for a <= a_mat
def f_coll1(a, abar, l, x, N, tol):
    global phi0
    y = np.array([phi0]*(N+1))
    sol = solve_newton_coll1(a, abar, l, x, y, N)
    norm = np.linalg.norm(sol)
    for i in range(N+1):
        y[i] += sol[i]
    it = 0
    while norm > tol:
        it += 1
        sol = solve_newton_coll1(a, abar, l, x, y, N)
        norm = np.linalg.norm(sol)
        for i in range(N+1):
            y[i] += sol[i]
    return y[:N]

#approximation of f by collocation at the initial value (l,x), for a > a_mat
def f_coll2(a, abar, start, l, x, N, tol):
    global phi0
    y = np.array([start]*(N+1))
    sol = solve_newton_coll2(a, abar, l, x, y, N)
    norm = np.linalg.norm(sol)
    for i in range(N+1):
        y[i] += sol[i]
    it = 0
    while norm > tol:
        it += 1
        sol = solve_newton_coll2(a, abar, l, x, y, N)
        norm = np.linalg.norm(sol)
        for i in range(N+1):
            y[i] += sol[i]
    return y[:N]

#computation of residual
def G(x, l, Phi, age):
    global phi0
    nnodes = len(Phi)
    nodes = (1 - age)*chebyshev_extrema_01(nnodes-1) + age
    N = sum([len(phi) for phi in Phi])
    out = np.zeros(N+2)
    fin = np.array([])
    for phi in range(nnodes-1):
        a = nodes[phi]
        nvars = len(Phi[phi])/2 + 1
        D = chebdiff_01(nvars-1, a-age)
        Dage = chebdiff_01(nvars-1, age)
        diffs = D.dot(Phi[phi][:nvars])
        diffsage = Dage.dot(np.append(Phi[phi][nvars-1:], phi0))
        expra = ((Phi[phi][:nvars] - phi0*0.5)/(x*(l**2 + 1)*(a + 1)) + 1/(age+1))
        exprb = ((np.append(Phi[phi][nvars-1:], phi0) - phi0)/(x*(l**2 + 1)) - 0.5)
        for i in range(nvars-1):
            out[2*phi*(nvars-1) + i] = diffs[i] + x*(l**2 + 1)*(a + 1)*expra[i]**2
            out[2*phi*(nvars-1) + i+nvars-1] = diffsage[i] + x*(l**2 + 1)*exprb[i]**2
        fin = np.append(fin, Phi[phi][0])
    phi = nnodes-1
    a = nodes[phi]
    nvars = len(Phi[phi]) + 1
    Dage = chebdiff_01(nvars-1, age)
    diffsage = Dage.dot(np.append(Phi[phi], phi0))
    expr = ((np.append(Phi[phi], phi0) - phi0)/(x*(l**2 + 1)) - 0.5)
    for i in range(nvars-1):
        out[2*phi*(nvars-1) + i] = diffsage[i] + x*(l**2 + 1)*expr[i]**2
    fin = np.append(fin, Phi[phi][0])
    weights = (1 - age)*clencurt_01(nnodes-1)
    out[N] = fin.dot(weights)
    out[N+1] = Phi[nnodes-1][0] - phi0*0.5
    return out

#auxiliary function for nullspace of a matrix
def null(A, eps=1e-9):
    u, s, vh = np.linalg.svd(A)
    null_mask = (s <= eps)
    null_space = np.compress(null_mask, vh, axis=0)
    return np.transpose(null_space)

def start_direction(x, l, Phi, age):
    global phi0
    nnodes = len(Phi)
    nodes = chebyshev_extrema_01(nnodes-1)
    weights = clencurt_01(nnodes-1)
    N = sum([len(phi) for phi in Phi])
    A = np.zeros((N+3, N+3))
    tot = 0
    I = np.eye(nnodes-1)
    Dlagrange = np.zeros(nnodes-1)
    for phi in range(nnodes-1):
        a = (1 - age)*nodes[phi] + age
        nvars = len(Phi[phi])/2 + 1
        expra = ((Phi[phi][:nvars] - phi0*0.5)/(x*(l**2 + 1)*(a + 1)) + 1/(age+1))
        exprb = ((np.append(Phi[phi][nvars-1:], phi0) - phi0)/(x*(l**2 + 1)) - 0.5)
        dadg = 1/((a + 1)*(l**2 + 1))
        dadf = dadg/x
        dbdg = 1/(l**2 + 1)
        dbdf = dbdg/x
        dadx = -(Phi[phi][:nvars] - phi0*0.5)/(x**2*(l**2 + 1)*(a + 1))
        dbdx = -(np.append(Phi[phi][nvars-1:], phi0) - phi0)/(x**2*(l**2 + 1))
        dadl = -2*l*(Phi[phi][:nvars] - phi0*0.5)/(x*(l**2 + 1)**2*(a + 1))
        dbdl = -2*l*(np.append(Phi[phi][nvars-1:], phi0) - phi0)/(x*(l**2 + 1)**2)
        dda = -(Phi[phi][:nvars] - phi0*0.5)/(x*(l**2 + 1)*(a + 1)**2)
        D = chebdiff_01(nvars-1, a-age)
        Dage = chebdiff_01(nvars-1, age)
        diffs = D.dot(Phi[phi][:nvars])
        diffsage = Dage.dot(np.append(Phi[phi][nvars-1:], phi0))
        for i in range(tot, tot + nvars - 1):
            for j in range(tot, tot + nvars):
                A[i,j] = D[i-tot,j-tot]
            for j in range(tot, tot + nvars - 1):
                A[i+nvars-1,j+nvars-1] += Dage[i-tot,j-tot]
            A[i, i] += 2*(a + 1)*x*(l**2 + 1)*dadf*expra[i-tot]
            A[i+nvars-1, i+nvars-1] += 2*x*(l**2 + 1)*dbdf*exprb[i-tot]
            A[i,N+1] += (a + 1)*(l**2 + 1)*expra[i-tot]*(expra[i-tot] + 2*x*dadx[i-tot])
            A[i+nvars-1,N+1] += (l**2 + 1)*exprb[i-tot]*(exprb[i-tot] + 2*x*dbdx[i-tot])
            A[i,N+2] += (a + 1)*x*expra[i-tot]*(2*l*expra[i-tot] + 2*(l**2 + 1)*dadl[i-tot])
            A[i+nvars-1,N+2] += x*exprb[i-tot]*(2*l*exprb[i-tot] + 2*(l**2 + 1)*dbdl[i-tot])
            A[i,N] += (nodes[phi])*diffs[i-tot]/(a-age) + x*(l**2 + 1)*(1 - nodes[phi])*expra[i-tot]*(2*(a + 1)*dda[i-tot] + expra[i-tot]) - 2*(a + 1)*x*(l**2 + 1)*expra[i-tot]/(1 + age)**2
            A[i+nvars-1,N] -= diffsage[i-tot]/age
        A[N,tot] = (1 - age)*weights[phi]
        tot += 2*(nvars - 1)
    phi = nnodes-1
    a = (1 - age)*nodes[phi] + age
    expr = (np.append(Phi[phi], phi0) - phi0)/(x*(l**2 + 1)) - 0.5
    ddg = 1/(l**2 + 1)
    ddf = ddg/x
    ddx = -(np.append(Phi[phi], phi0) - phi0)/(x**2*(l**2 + 1))
    ddl = -2*l*(np.append(Phi[phi], phi0) - phi0)/(x*(l**2 + 1)**2)
    nvars = len(Phi[phi]) + 1
    D = chebdiff_01(nvars-1, a)
    diffs = D.dot(np.append(Phi[phi], phi0))
    for i in range(tot, tot + nvars - 1):
        for j in range(tot, tot + nvars -1):
            A[i,j] += D[i-tot,j-tot]
        A[i, i] += 2*x*(l**2 + 1)*ddf*expr[i-tot]
        A[i,N+1] += (l**2 + 1)*expr[i-tot]*(expr[i-tot] + 2*x*ddx[i-tot])
        A[i,N+2] += x*expr[i-tot]*(2*l*expr[i-tot] + 2*(l**2 + 1)*ddl[i-tot])
        A[i,N] += (nodes[phi]-1)*diffs[i-tot]/a
    A[N,tot] = (1 - age)*weights[phi]
    A[N,N] = - sum([weights[p]*Phi[p][0] for p in range(nnodes)])
    A[N+1,2*(nvars-1)*(nnodes-1)] = 1.
    d = null(A)
    return d

#computation of Jacobian and update of the variables
def solve_newton(x, l, Phi, age, arr, d):
    global phi0
    nnodes = len(Phi)
    nodes = chebyshev_extrema_01(nnodes-1)
    weights = clencurt_01(nnodes-1)
    N = sum([len(phi) for phi in Phi])
    A = np.zeros((N+3, N+3))
    tot = 0
    I = np.eye(nnodes-1)
    Dlagrange = np.zeros(nnodes-1)
    for phi in range(nnodes-1):
        a = (1 - age)*nodes[phi] + age
        nvars = len(Phi[phi])/2 + 1
        expra = ((Phi[phi][:nvars] - phi0*0.5)/(x*(l**2 + 1)*(a + 1)) + 1/(age+1))
        exprb = ((np.append(Phi[phi][nvars-1:], phi0) - phi0)/(x*(l**2 + 1)) - 0.5)
        dadg = 1/((a + 1)*(l**2 + 1))
        dadf = dadg/x
        dbdg = 1/(l**2 + 1)
        dbdf = dbdg/x
        dadx = -(Phi[phi][:nvars] - phi0*0.5)/(x**2*(l**2 + 1)*(a + 1))
        dbdx = -(np.append(Phi[phi][nvars-1:], phi0) - phi0)/(x**2*(l**2 + 1))
        dadl = -2*l*(Phi[phi][:nvars] - phi0*0.5)/(x*(l**2 + 1)**2*(a + 1))
        dbdl = -2*l*(np.append(Phi[phi][nvars-1:], phi0) - phi0)/(x*(l**2 + 1)**2)
        dda = -(Phi[phi][:nvars] - phi0*0.5)/(x*(l**2 + 1)*(a + 1)**2)
        D = chebdiff_01(nvars-1, a-age)
        Dage = chebdiff_01(nvars-1, age)
        diffs = D.dot(Phi[phi][:nvars])
        diffsage = Dage.dot(np.append(Phi[phi][nvars-1:], phi0))
        for i in range(tot, tot + nvars - 1):
            for j in range(tot, tot + nvars):
                A[i,j] = D[i-tot,j-tot]
            for j in range(tot, tot + nvars - 1):
                A[i+nvars-1,j+nvars-1] += Dage[i-tot,j-tot]
            A[i, i] += 2*(a + 1)*x*(l**2 + 1)*dadf*expra[i-tot]
            A[i+nvars-1, i+nvars-1] += 2*x*(l**2 + 1)*dbdf*exprb[i-tot]
            A[i,N+1] += (a + 1)*(l**2 + 1)*expra[i-tot]*(expra[i-tot] + 2*x*dadx[i-tot])
            A[i+nvars-1,N+1] += (l**2 + 1)*exprb[i-tot]*(exprb[i-tot] + 2*x*dbdx[i-tot])
            A[i,N+2] += (a + 1)*x*expra[i-tot]*(2*l*expra[i-tot] + 2*(l**2 + 1)*dadl[i-tot])
            A[i+nvars-1,N+2] += x*exprb[i-tot]*(2*l*exprb[i-tot] + 2*(l**2 + 1)*dbdl[i-tot])
            A[i,N] += (nodes[phi])*diffs[i-tot]/(a-age) + x*(l**2 + 1)*(1 - nodes[phi])*expra[i-tot]*(2*(a + 1)*dda[i-tot] + expra[i-tot]) - 2*(a + 1)*x*(l**2 + 1)*expra[i-tot]/(1 + age)**2
            A[i+nvars-1,N] -= diffsage[i-tot]/age
        A[N,tot] = (1 - age)*weights[phi]
        tot += 2*(nvars - 1)
    phi = nnodes-1
    a = (1 - age)*nodes[phi] + age
    expr = (np.append(Phi[phi], phi0) - phi0)/(x*(l**2 + 1)) - 0.5
    ddg = 1/(l**2 + 1)
    ddf = ddg/x
    ddx = -(np.append(Phi[phi], phi0) - phi0)/(x**2*(l**2 + 1))
    ddl = -2*l*(np.append(Phi[phi], phi0) - phi0)/(x*(l**2 + 1)**2)
    nvars = len(Phi[phi]) + 1
    D = chebdiff_01(nvars-1, a)
    diffs = D.dot(np.append(Phi[phi], phi0))
    for i in range(tot, tot + nvars - 1):
        for j in range(tot, tot + nvars -1):
            A[i,j] += D[i-tot,j-tot]
        A[i, i] += 2*x*(l**2 + 1)*ddf*expr[i-tot]
        A[i,N+1] += (l**2 + 1)*expr[i-tot]*(expr[i-tot] + 2*x*ddx[i-tot])
        A[i,N+2] += x*expr[i-tot]*(2*l*expr[i-tot] + 2*(l**2 + 1)*ddl[i-tot])
        A[i,N] += (nodes[phi]-1)*diffs[i-tot]/a
    A[N,tot] = (1 - age)*weights[phi]
    A[N,N] = - sum([weights[phi]*Phi[phi][0] for phi in range(nnodes)])
    A[N+1,2*(nvars-1)*(nnodes-1)] = 1.
    for j in range(N+3):
        A[N+2,j] = d[j]
    sol = np.linalg.solve(A, arr)
    return sol

#Newton's method, single continuation step
def pseudo_alc(u0, lam0, Phi, a0, ds, d, tol):
    N = len(d) - 3
    if d[N+2] < 0:
        ds = -ds
    new_u = u0 + ds*d[N+1]
    new_lam = lam0 + ds*d[N+2]
    new_a = a0 + ds*d[N]
    new_Phi = copy.deepcopy(Phi)
    nnodes = len(Phi)
    tot = 0
    for phi in range(nnodes):
        nvars = len(Phi[phi])
        for i in range(tot, tot + nvars):
            new_Phi[phi][i-tot] += ds*d[i]
        tot += nvars
    norm = 1.
    it = 0
    while norm > tol and it < 20:
        it += 1
        diff = np.zeros(N)
        tot = 0
        for phi in range(nnodes):
            nvars = len(Phi[phi])
            for i in range(tot, tot + nvars):
                diff[i] = new_Phi[phi][i-tot] - Phi[phi][i-tot]
            tot += nvars
        b = -np.append(G(new_u, new_lam, new_Phi, new_a), (new_u - u0)*d[N+1] + (new_lam - lam0)*d[N+2] + (new_a - a0)*d[N] + diff.dot(d[:N]) - ds)
        sol = solve_newton(new_u, new_lam, new_Phi, new_a, b, d)
        norm = np.linalg.norm(sol)
        new_u += sol[N+1]
        new_lam += sol[N+2]
        new_a += sol[N]
        tot = 0
        for phi in range(nnodes):
            nvars = len(Phi[phi])
            for i in range(tot, tot + nvars):
                new_Phi[phi][i-tot] += sol[i]
            tot += nvars
    new_d = solve_newton(new_u, new_lam, new_Phi, new_a, np.append(np.zeros(N+2), 1.), d)
    n = np.linalg.norm(new_d)
    if n != 0:
        new_d = new_d/n
    if it == 20:
        return lam0, u0, Phi, a0, d, it
    return new_lam, new_u, new_Phi, new_a, new_d, it

#varying number of collocation points, fixed number of quadrature points
for ncoll in range(7,8):
    nquad = 10
    l = 1.
    u = exact(l)
    abar = 0.5
    test = [f_coll1(a, abar, l, u, ncoll-1, 1e-12) for a in (1 - abar)*chebyshev_extrema_01(nquad) + abar]
    nodes = (1 - abar)*chebyshev_extrema_01(nquad)[:nquad] + abar
    test2 = [f_coll2(nodes[i], abar, test[i][0], l, u, ncoll-1, 1e-12) for i in range(len(nodes))]
    Phi = [np.concatenate((test2[i], test[i])) for i in range(len(nodes))]+[test[nquad]]
    d = start_direction(u, l, Phi, abar)
    ds = 0.02*ncoll
    start = time.time()
    #10 continuation steps
    l, u, Phi, a, d, it = pseudo_alc(u, l, Phi, abar, ds, d, 1e-13)
    while it == 20:
        ds = ds*0.5
        l, u, Phi, a, d, it = pseudo_alc(u, l, Phi, abar, ds, d, 1e-13)
    lambdas = np.array([l])
    us = np.array([u])
    its = np.array([it])
    if it < 3:
        ds = ds*1.3

    for i in range(9):
        l, u, Phi, a, d, it = pseudo_alc(u, l, Phi, a, ds, d, 1e-13)
        while it == 20:
            ds *= 0.5
            l, u, Phi, a, d, it = pseudo_alc(u, l, Phi, a, ds, d, 1e-13)
        lambdas = np.append(lambdas, l)
        us = np.append(us, u)
        its = np.append(its, it)
        if it < 3:
            ds = ds*1.3

    elapsed = time.time() - start
    print 'Elapsed time =', elapsed

    errs = []
    for l1, u1, it in zip(lambdas, us, its):
        print l1, abs(u1-exact(l1)), it
        errs.append(abs(u1-exact(l1)))
