'''
Example from section 6.1 in [1]

This script performs internal continuation of the curve x(lambda) defined by

int_{0}^{1}f(a,x,lambda) d{a}=0

where f(a,x,lambda)=varphi(a;a,x,lambda) is defined by

{varphi'(alpha;a,x,lambda)=g(varphi(alpha;a,x,lambda),a,x,lambda), alpha in [0,a], \\
varphi(0;a,x,lambda)=varphi_0.

with g(varphi(alpha;a,x,lambda),a,x,lambda)=lambda varphi(alpha;a,x,lambda) + 2xe^{-lambda a}
and varphi_0=1.

Copyright (C) 2019 Alessia Ando

[1] A. Ando, D. Breda, F. Scarabel, Numerical continuation and delay equations: A novel approach
for complex models of structured populations, Discrete Contin. Dyn. Syst. Ser. S, DOI: 110.3934/dcdss.2020165
'''

import numpy as np
from scipy.fftpack import ifft
import time
import copy

#exact expression of x(lambda) to monitor true error
def exact(t):
    if t == 0:
        return -2.
    else:
        return t*(1-np.exp(t))*0.5/(t + np.exp(-t) - 1)

#collocation on chebyshev extrema in [0,a]
def chebyshev_extrema_01(n):
    return np.array([0.5*(1 + np.cos(k*np.pi/n)) for k in range(n+1)])

#differentiation matrix for chebyshev nodes in [0,a]
def chebdiff_01(n, a):
    f = 1/a
    D = np.zeros((n+1, n+1))
    x = chebyshev_extrema_01(n)
    D[0,0] = (2*n**2 + 1)*f/3.
    D[n,n] = -(2*n**2 + 1)*f/3.
    D[0,n] = (-1)**n*f
    D[n,0] = -(-1)**n*f
    for j in range(1, n):
        D[0,j] = 2*(-1)**j*f/(1 - x[j])
        D[n,j] = -2*(-1)**(n+j)*f/x[j]
    for i in range(1, n):
        D[i,0] = -0.5*(-1)**i*f/(1 - x[i])
        D[i,n] = 0.5*(-1)**(n+i)*f/x[i]
    for i in range(1, n):
        for j in range(1, i):
            D[i,j] = (-1)**(i+j)*f/(x[i] - x[j])
            D[j,i] = -(-1)**(i+j)*f/(x[i] - x[j])
        D[i, i] = -(2*x[i] - 1)*f/(1 - (2*x[i] - 1)**2)
    return D

#Clenshaw-Curtis weights scaled to interval [0,1]
def clencurt_01(n):
    C = np.zeros((n+1,2))
    k = 2*(1+np.arange(np.floor(n/2)))
    C[::2,0] = 2/np.hstack((1, 1-k*k))
    C[1,1] = -n
    V = np.vstack((C,np.flipud(C[1:n,:])))
    F = np.real(ifft(V, n=None, axis=0))
    w = np.hstack((F[0,0],2*F[1:n,0],F[n,0]))
    return 0.5*w

#approximation of f by collocation at the initial value (l,x)
def f_coll(a, l, x, N):
    coll_nodes = [a*n for n in chebyshev_extrema_01(N)]
    b = 2*x*np.exp(-l*a)*np.ones(N)
    b = np.append(b,1)
    M = chebdiff_01(N, a)
    M[N,N] = 1.
    for j in range(N):
        M[N,j] = 0.
        M[j,j] -= l
    phi = np.linalg.solve(M, b)
    return phi[:N]

#computation of residual
def G(u, t, Phi):
    nnodes = len(Phi) + 1
    nodes = chebyshev_extrema_01(nnodes-1)
    out = np.array([])
    fin = np.array([])
    for phi in range(nnodes - 1):
        a = nodes[phi]
        nvars = len(Phi[phi]) + 1
        D = chebdiff_01(nvars-1, a)
        diffs = D.dot(np.append(Phi[phi],1))
        for i in range(nvars-1):
            val = Phi[phi][i]
            out = np.append(out, diffs[i] - t*val - 2*u*np.exp(-t*a))
        fin = np.append(fin, Phi[phi][0])
    fin = np.append(fin,1)
    weights = clencurt_01(nnodes-1)
    out = np.append(out, fin.dot(weights))
    return out

#auxiliary function for nullspace of a matrix
def null(A, eps=1e-9):
    u, s, vh = np.linalg.svd(A)
    null_mask = (s <= eps)
    null_space = np.compress(null_mask, vh, axis=0)
    return np.transpose(null_space)

def start_direction(u, t, Phi):
    nnodes = len(Phi) + 1
    nodes = chebyshev_extrema_01(nnodes-1)
    weights = clencurt_01(nnodes-1)
    N = sum([len(phi) for phi in Phi])
    A = np.zeros((N+2,N+2))
    tot = 0
    for phi in range(nnodes - 1):
        a = nodes[phi]
        nvars = len(Phi[phi]) + 1
        D = chebdiff_01(nvars-1, a)
        for i in range(tot, tot + nvars - 1):
            for j in range(tot, tot + nvars - 1):
                A[i,j] = D[i-tot,j-tot]
            A[i,i] -= t
            A[i,N] = -2*np.exp(-t*a)
            A[i,N+1] = -Phi[phi][i-tot] + a*2*u*np.exp(-t*a)
        A[N,tot] = weights[phi]
        tot += nvars - 1
    d = null(A)
    return d

#computation of Jacobian and update of the variables
def solve_newton(u, t, Phi, arr, d):
    nnodes = len(Phi) + 1
    nodes = chebyshev_extrema_01(nnodes-1)
    weights = clencurt_01(nnodes-1)
    N = sum([len(phi) for phi in Phi])
    A = np.zeros((N+2,N+2))
    tot = 0
    for phi in range(nnodes - 1):
        a = nodes[phi]
        nvars = len(Phi[phi]) + 1
        D = chebdiff_01(nvars-1, a)
        for i in range(tot, tot + nvars - 1):
            for j in range(tot, tot + nvars - 1):
                A[i,j] = D[i-tot,j-tot]
            A[i,i] -= t
            A[i,N] = -2*np.exp(-t*a)
            A[i,N+1] = -Phi[phi][i-tot] + a*2*u*np.exp(-t*a)
        A[N,tot] = weights[phi]
        tot += nvars - 1
    for j in range(N+2):
        A[N+1,j] = d[j]
    sol = np.linalg.solve(A, arr)
    return sol

#Newton's method, single continuation step
def pseudo_alc(u0, lam0, Phi, ds, d, tol):
    N = len(d) - 2
    if d[N+1] < 0:
        ds = -ds
    new_u = u0 + ds*d[N]
    new_lam = lam0 + ds*d[N+1]
    tot = 0
    new_Phi = copy.deepcopy(Phi)
    nnodes = len(Phi)
    for phi in range(nnodes):
        nvars = len(Phi[phi])
        for i in range(tot, tot + nvars):
            new_Phi[phi][i-tot] += ds*d[i]
        tot += nvars
    b = -np.append(G(new_u, new_lam, new_Phi), 0)
    sol = solve_newton(new_u, new_lam, new_Phi, b, d)
    norm = np.linalg.norm(sol)
    new_u += sol[N]
    new_lam += sol[N+1]
    tot = 0
    for phi in range(nnodes):
        nvars = len(Phi[phi])
        for i in range(tot, tot + nvars):
            new_Phi[phi][i-tot] += sol[i]
        tot += nvars
    it = 0
    while norm > tol and it < 20:
        it += 1
        diff = np.zeros(N)
        tot = 0
        for phi in range(nnodes):
            nvars = len(Phi[phi])
            for i in range(tot, tot + nvars):
                diff[i] = new_Phi[phi][i-tot] - Phi[phi][i-tot]
            tot += nvars
        b = -np.append(G(new_u, new_lam, new_Phi), (new_u - u0)*d[N] + (new_lam - lam0)*d[N+1] + diff.dot(d[:N]) - ds)
        sol = solve_newton(new_u, new_lam, new_Phi, b, d)
        norm = np.linalg.norm(sol)
        new_u += sol[N]
        new_lam += sol[N+1]
        tot = 0
        for phi in range(nnodes):
            nvars = len(Phi[phi])
            for i in range(tot, tot + nvars):
                new_Phi[phi][i-tot] += sol[i]
            tot += nvars
    new_d = solve_newton(new_u, new_lam, new_Phi, np.append(np.zeros(N+1), 1.), d)
    n = np.linalg.norm(new_d)
    if n != 0:
        new_d = new_d/n
    if it == 20:
        return lam0, u0, Phi, d, it
    return new_lam, new_u, new_Phi, new_d, it

#fixed number of collocation points, varying number of quadrature points
for nquad in range(5, 6):
    l = 1.
    u = exact(l)
    ncoll = 10
    errs = []
    Phi = [f_coll(a, l, u, ncoll) for a in chebyshev_extrema_01(nquad)[:-1]]
    ds = 0.08*nquad
    d = start_direction(u, l, Phi)
    start = time.time()
    
    #10 continuation steps
    l, u, Phi, d, it = pseudo_alc(u, l, Phi, ds, d, 1e-14)
    while it == 20:
        ds *= 0.5
        l, u, Phi, d, it = pseudo_alc(u, l, Phi, ds, d, 1e-14)
    lambdas = np.array([l])
    us = np.array([u])
    its = np.array([it])
    if it < 3:
        ds *= 1.3
    
    for i in range(9):
        l, u, Phi, d, it = pseudo_alc(u, l, Phi, ds, d, 1e-14)
        while it == 20:
            ds *= 0.5
            l, u, Phi, d, it = pseudo_alc(u, l, Phi, ds, d, 1e-14)
        lambdas = np.append(lambdas, l)
        us = np.append(us, u)
        its = np.append(its, it)
        if it < 3:
            ds *= 1.3

    elapsed = time.time() - start
    print 'Elapsed time =', elapsed

    for l1, u1, it in zip(lambdas, us, its):
        print l1, abs(u1-exact(l1)), it
        errs.append(abs(u1-exact(l1)))

