'''
Example from section 6.2 in [1]

This script performs internal continuation of the curve x(lambda) defined by

int_{a_mat}^{1}f(a,x,lambda) d{a}=0

where f(a,x,lambda)=varphi(a;a,x,lambda) is defined by

{varphi'(alpha;a,x,lambda)=g(varphi(alpha;a,x,lambda),a,x,lambda), alpha in [0,a], \\
varphi(0;a,x,lambda)=varphi_0.

with g(varphi(alpha;a,x,lambda),a,x,lambda)=-(a + 1)x(lambda^2 + 2)({varphi(alpha;a,x,lambda) - varphi_0}/{x(lambda^2 + 2)(a + 1)} + 1)^2

and a_mat is defined by f(a_mat,x,lambda)=varphi_0-1.

Copyright (C) 2019 Alessia Ando

[1] A. Ando, D. Breda, F. Scarabel, Numerical continuation and delay equations: A novel approach
for complex models of structured populations, Discrete Contin. Dyn. Syst. Ser. S, DOI: 110.3934/dcdss.2020165
'''

import numpy as np
from scipy.fftpack import ifft
import time
import copy

#definition of varphi_0
phi0 = 1.5

#exact expression of x(lambda) to monitor true error
def exact(l):
    global phi0
    return (2*phi0 - 1)/(l**2 + 2)

#collocation on chebyshev extrema in [0,a]
def chebyshev_extrema_01(n):
    return np.array([0.5*(1 + np.cos(k*np.pi/n)) for k in range(n+1)])

#differentiation matrix for chebyshev nodes in [0,a]
def chebdiff_01(n, a):
    f = 1/a
    D = np.zeros((n+1, n+1))
    x = chebyshev_extrema_01(n)
    D[0,0] = (2*n**2 + 1)*f/3.
    D[n,n] = -(2*n**2 + 1)*f/3.
    D[0,n] = (-1)**n*f
    D[n,0] = -(-1)**n*f
    for j in range(1, n):
        D[0,j] = 2*(-1)**j*f/(1 - x[j])
        D[n,j] = -2*(-1)**(n+j)*f/x[j]
    for i in range(1, n):
        D[i,0] = -0.5*(-1)**i*f/(1 - x[i])
        D[i,n] = 0.5*(-1)**(n+i)*f/x[i]
    for i in range(1, n):
        for j in range(1, i):
            D[i,j] = (-1)**(i+j)*f/(x[i] - x[j])
            D[j,i] = -(-1)**(i+j)*f/(x[i] - x[j])
        D[i, i] = -(2*x[i] - 1)*f/(1 - (2*x[i] - 1)**2)
    return D

#Clenshaw-Curtis weights scaled to interval [0,1]
def clencurt_01(n):
    C = np.zeros((n+1,2))
    k = 2*(1+np.arange(np.floor(n/2)))
    C[::2,0] = 2/np.hstack((1, 1-k*k))
    C[1,1] = -n
    V = np.vstack((C,np.flipud(C[1:n,:])))
    F = np.real(ifft(V, n=None, axis=0))
    w = np.hstack((F[0,0],2*F[1:n,0],F[n,0]))
    return 0.5*w

#auxiliary function for f_coll below
def solve_newton_coll(a, l, x, y, N):
    global phi0
    M = chebdiff_01(N,a)
    M[N,N] = 1.
    for j in range(N):
        M[N,j] = 0.
    expr = ((y - phi0)/(x*(l**2 + 2)*(a + 1)) + 1)
    func = M.dot(y) + x*(l**2 + 2)*(a + 1)*expr**2
    func[N] = 0
    A = np.copy(M)
    ddg = 1/((a + 1)*(l**2 + 2))
    ddf = ddg/x
    for i in range(N):
        A[i, i] += 2*(a + 1)*x*(l**2 + 2)*ddf*expr[i]
    sol = np.linalg.solve(A, -func)
    return sol

#approximation of f by collocation at the initial value (l,x)
def f_coll(a, l, x, N, tol):
    global phi0
    y = np.array([phi0]*(N+1))
    sol = solve_newton_coll(a, l, x, y, N)
    norm = np.linalg.norm(sol)
    for i in range(N+1):
        y[i] += sol[i]
    it = 0
    while norm > tol:
        it += 1
        sol = solve_newton_coll(a, l, x, y, N)
        norm = np.linalg.norm(sol)
        for i in range(N+1):
            y[i] += sol[i]
    return y[:N]

#computation of residual
def G(x, l, Phi, age):
    global phi0
    nnodes = len(Phi)
    nodes = (1 - age)*chebyshev_extrema_01(nnodes-1) + age
    N = sum([len(phi) for phi in Phi])
    out = np.zeros(N+2)
    fin = np.array([])
    for phi in range(nnodes):
        a = nodes[phi]
        nvars = len(Phi[phi]) + 1
        D = chebdiff_01(nvars-1, a)
        diffs = D.dot(np.append(Phi[phi], phi0))
        expr = ((Phi[phi] - phi0)/(x*(l**2 + 2)*(a + 1)) + 1)
        for i in range(nvars-1):
            out[phi*(nvars-1) + i] = diffs[i] + x*(l**2 + 2)*(a + 1)*expr[i]**2
        fin = np.append(fin, Phi[phi][0])
    weights = (1 - age)*clencurt_01(nnodes-1)
    out[N] = fin.dot(weights)
    out[N+1] = Phi[nnodes-1][0] - phi0 + 1
    return out

#auxiliary function for nullspace of a matrix
def null(A, eps=1e-9):
    u, s, vh = np.linalg.svd(A)
    null_mask = (s <= eps)
    null_space = np.compress(null_mask, vh, axis=0)
    return np.transpose(null_space)

def start_direction(x, l, Phi, age):
    global phi0
    nnodes = len(Phi)
    nodes = chebyshev_extrema_01(nnodes-1)
    weights = clencurt_01(nnodes-1)
    N = sum([len(phi) for phi in Phi])
    A = np.zeros((N+3, N+3))
    tot = 0
    I = np.eye(nnodes-1)
    Dlagrange = np.zeros(nnodes-1)
    for phi in range(nnodes):
        a = (1 - age)*nodes[phi] + age
        expr = ((Phi[phi] - phi0)/(x*(l**2 + 2)*(a + 1)) + 1)
        ddg = 1/((a + 1)*(l**2 + 2))
        ddf = ddg/x
        ddx = -(Phi[phi] - phi0)/(x**2*(l**2 + 2)*(a + 1))
        ddl = -2*l*(Phi[phi] - phi0)/(x*(l**2 + 2)**2*(a + 1))
        dda = -(Phi[phi] - phi0)/(x*(l**2 + 2)*(a + 1)**2)
        nvars = len(Phi[phi]) + 1
        D = chebdiff_01(nvars-1, a)
        diffs = D.dot(np.append(Phi[phi], phi0))
        for i in range(tot, tot + nvars - 1):
            for j in range(tot, tot + nvars - 1):
                A[i,j] = D[i-tot,j-tot]
            A[i, i] += 2*(a + 1)*x*(l**2 + 2)*ddf*expr[i-tot]
            A[i,N+1] += (a + 1)*(l**2 + 2)*expr[i-tot]*(expr[i-tot] + 2*x*ddx[i-tot])
            A[i,N+2] += (a + 1)*x*expr[i-tot]*(2*l*expr[i-tot] + 2*(l**2 + 2)*ddl[i-tot])
            A[i,N] += (nodes[phi] - 1)*diffs[i-tot]/a + x*(l**2 + 2)*(1 - nodes[phi])*expr[i-tot]*(2*(a + 1)*dda[i-tot] + expr[i-tot])
        A[N,tot] = (1 - age)*weights[phi]
        tot += (nvars - 1)
    A[N,N] = - sum([weights[phi]*Phi[phi][0] for phi in range(nnodes)])
    A[N+1,(nvars-1)*(nnodes-1)] = 1.
    d = null(A)
    return d

#computation of Jacobian and update of the variables
def solve_newton(x, l, Phi, age, arr, d):
    global phi0
    nnodes = len(Phi)
    nodes = chebyshev_extrema_01(nnodes-1)
    weights = clencurt_01(nnodes-1)
    N = sum([len(phi) for phi in Phi])
    A = np.zeros((N+3, N+3))
    tot = 0
    I = np.eye(nnodes-1)
    Dlagrange = np.zeros(nnodes-1)
    for phi in range(nnodes):
        a = (1 - age)*nodes[phi] + age
        expr = ((Phi[phi] - phi0)/(x*(l**2 + 2)*(a + 1)) + 1)
        ddg = 1/((a + 1)*(l**2 + 2))
        ddf = ddg/x
        ddx = -(Phi[phi] - phi0)/(x**2*(l**2 + 2)*(a + 1))
        ddl = -2*l*(Phi[phi] - phi0)/(x*(l**2 + 2)**2*(a + 1))
        dda = -(Phi[phi] - phi0)/(x*(l**2 + 2)*(a + 1)**2)
        nvars = len(Phi[phi]) + 1
        D = chebdiff_01(nvars-1, a)
        diffs = D.dot(np.append(Phi[phi], phi0))
        for i in range(tot, tot + nvars - 1):
            for j in range(tot, tot + nvars - 1):
                A[i,j] = D[i-tot,j-tot]
            A[i, i] += 2*(a + 1)*x*(l**2 + 2)*ddf*expr[i-tot]
            A[i,N+1] += (a + 1)*(l**2 + 2)*expr[i-tot]*(expr[i-tot] + 2*x*ddx[i-tot])
            A[i,N+2] += (a + 1)*x*expr[i-tot]*(2*l*expr[i-tot] + 2*(l**2 + 2)*ddl[i-tot])
            A[i,N] += (nodes[phi] - 1)*diffs[i-tot]/a + x*(l**2 + 2)*(1 - nodes[phi])*expr[i-tot]*(2*(a + 1)*dda[i-tot] + expr[i-tot])
        A[N,tot] = (1 - age)*weights[phi]
        tot += (nvars - 1)
    A[N,N] = - sum([weights[phi]*Phi[phi][0] for phi in range(nnodes)])
    A[N+1,(nvars-1)*(nnodes-1)] = 1.
    for j in range(N+3):
        A[N+2,j] = d[j]
    sol = np.linalg.solve(A, arr)
    return sol

#Newton's method, single continuation step
def pseudo_alc(u0, lam0, Phi, a0, ds, d, tol):
    N = len(d) - 3
    if d[N+2] < 0:
        ds = -ds
    new_u = u0 + ds*d[N+1]
    new_lam = lam0 + ds*d[N+2]
    new_a = a0 + ds*d[N]
    new_Phi = copy.deepcopy(Phi)
    nnodes = len(Phi)
    tot = 0
    for phi in range(nnodes):
        nvars = len(Phi[phi])
        for i in range(tot, tot + nvars):
            new_Phi[phi][i-tot] += ds*d[i]
        tot += nvars
    norm = 1.
    it = 0
    while norm > tol and it < 20:
        it += 1
        diff = np.zeros(N)
        tot = 0
        for phi in range(nnodes):
            nvars = len(Phi[phi])
            for i in range(tot, tot + nvars):
                diff[i] = new_Phi[phi][i-tot] - Phi[phi][i-tot]
            tot += nvars
        b = -np.append(G(new_u, new_lam, new_Phi, new_a), (new_u - u0)*d[N+1] + (new_lam - lam0)*d[N+2] + (new_a - a0)*d[N] + diff.dot(d[:N]) - ds)
        sol = solve_newton(new_u, new_lam, new_Phi, new_a, b, d)
        norm = np.linalg.norm(sol)
        new_u += sol[N+1]
        new_lam += sol[N+2]
        new_a += sol[N]
        tot = 0
        for phi in range(nnodes):
            nvars = len(Phi[phi])
            for i in range(tot, tot + nvars):
                new_Phi[phi][i-tot] += sol[i]
            tot += nvars
    new_d = solve_newton(new_u, new_lam, new_Phi, new_a, np.append(np.zeros(N+2), 1.), d)
    n = np.linalg.norm(new_d)
    if n != 0:
        new_d = new_d/n
    if it == 20:
        return lam0, u0, Phi, a0, d, it
    return new_lam, new_u, new_Phi, new_a, new_d, it

#varying number of collocation points, fixed number of quadrature points
for ncoll in range(10, 11):
    nquad = 5
    l = 1.
    u = exact(l)
    Phi = [f_coll(a, l, u, ncoll-1, 1e-12) for a in 0.5*chebyshev_extrema_01(nquad) + 0.5]
    d = start_direction(u, l, Phi, 0.5)
    ds = 0.015*ncoll
    start = time.time()
    #10 continuation steps
    l, u, Phi, a, d, it = pseudo_alc(u, l, Phi, 0.5, ds, d, 1e-13)
    while it == 20:
        ds = ds*0.5
        l, u, Phi, a, d, it = pseudo_alc(u, l, Phi, 0.5, ds, d, 1e-13)
    lambdas = np.array([l])
    us = np.array([u])
    its = np.array([it])
    if it < 3:
        ds = ds*1.3

    for i in range(9):
        l, u, Phi, a, d, it = pseudo_alc(u, l, Phi, a, ds, d, 1e-13)
        while it == 20:
            ds *= 0.5
            l, u, Phi, a, d, it = pseudo_alc(u, l, Phi, a, ds, d, 1e-13)
        lambdas = np.append(lambdas, l)
        us = np.append(us, u)
        its = np.append(its, it)
        if it < 3:
            ds = ds*1.3

    elapsed = time.time() - start
    print 'Elapsed time =', elapsed

    errs = []
    for l1, u1, it in zip(lambdas, us, its):
        print l1, abs(u1-exact(l1)), it
        errs.append(abs(u1-exact(l1)))

